/*
DL2 Monitor: ./pa/dl2pa.c
Copyright (C) 2002, Rolf Jger, Ulm (Germany)

See the file README for information on usage and redistribution
of this source code file, and for a DISCLAIMER OF ALL WARRANTIES.
*/

/*----------------------------------------------------------------------------
 * Datalink Monitor (dl2pa): Protocol Analysis (Main Process)
 *
 * The process provides for online/offline protocol analysis and, in online
 * mode, it is considered the peer of the DL2 Monitor process. Therefore, it
 * is called the DL2 Analyzer. The online operation is the default. In this 
 * mode, the Analyzer receives data from stdin and uses IPC via shared memory.
 * 
 * Online Mode 
 * The process initializes known packet analyzers, selects an actual analyzer
 * according to the settings in the Monitor data header read from input, and 
 * executes that analyzer until termed by SIGTERM or in reponse to some other
 * events. The result always will be written to stdout. The input is always 
 * received from stdin and considered a sequence of identical formatted packets. 
 * Thus, the selected analyzer doesn't change until explicitly told to do so.
 * Data are read until the process has been kill by SIGTERM or SIGINT.
 *
 * Offline Mode
 * The process behaves similarly, except using IPC and reading from stdin. 
 * Consequently, only one analyzer will be executed at all, and execution 
 * will terminate on end-of-file. 
 *
 * Selection of the actual analyzer depends on the socket-type, protocol, and
 * network interface as being understood by the Linux packet socket. Because
 * of the packet socket doesn't any filtering on which an analyzer could have
 * been setup, some experiments have led to some conjectures on the behaviour
 * of the packet socket. The current implementation of this module depends on
 * that and is close to the ethernet interface. Currently, there are three 
 * toplevel analyzers available and in use: paehter2, papppoe, and paip for
 * the Ether II, PPPoE, and IP packets, respectively. 
 *
 * The process must be instructed via an rcfile or command line arguments.
 * As usual, command line arguments will precede those of the rcfile. The
 * rcfile provides mainly for field selections and for the Monitor input file.
 * It is searched for in in the dl2mon-directory given in the environment by
 * DL2MON_DIR and then in the users HOME directory. Its basename is .dl2monrc.
 * The rcfile can also be specified via a command line argument, as well as
 * the Monitor input file (offline) and some field selections.
 *
 * Usage: dl2pa {options}
 *     local options:
 * -F{<flag>}    set some <flag> which normally is unset:
 *  o            force analyzer into offline mode
 * -o  <file>    write output to <file> instead to stdout (currently ignored)
 * -fo <format>  generate output according to <format>
 * -s  <p>:{<f>} set selections for packet <p> to {<f>}
 * -a  <p>:{<f>} add selections {<f>} for packet <p>
 * -r  <p>:{<f>} remove selections {<f>} for packet <p>
 * -vt           print known packet types and exit
 * -vf <vt>      print known fields for packet type <vt> and exit 
 *     from lib/monpa.h:
 * -C  <file>    use <file> instead of the rcfile
 *     offline relevant, only:
 * -i  <file>    read Monitor input from <file> (forces offline mode)
 * -fi <format>  consider Monitor input to be formed according to <format> 
 *
 * Implementation Notes:
 * -monfd: file descriptor of Monitor output
 * -shmid: shared memory identifier as returned by shmget()
 * -shmmonpa: local shared memory address as returned by shmat()
 * -terminate: set on SIGTERM and SIGINT events
 * -verbose: table of verbose flags derived from rcfile or command line
 * -verbose_set/add/rem: verbose flags derived from command line, only
 * -configuration: table of configuration references contained in library
 * -act_analyzer, act_verbose:
 *  these objects represent the sole and actual anaylzer. They will be derived
 *  from header in Monitor output.
 * -time_offset:
 *  denotes the absolute timestamp of the packet processed first;
 *  timestamps will be output relatively, starting with 0 for the first packet.
 * 
 * Author: Rolf Jger,Ulm
 *----------------------------------------------------------------------------
 * YYYYMMDD  History
 * 20020501  DL2_VERSION: package version introduced.
 * 20020427  -vt, -vf, -s, -a, -r options implemented.
 * 20020414  use of dl2pa_libinit().
 * 20020411  online: input read from stdin, child of Monitor.
 * 20020410  shared memory usage introduced.
 * 20020410  packet-length, default: supported here by PA_VM*
 * 20020408  act_analyzer, act_verbose from  Dl2_Monheader. 
 * 20020407  command line args, rcfile usage introduced. 
 * 20020405  get_pconfig: bug fixed, there may be 0-entries.
 * 20020330  created from scratch
 * $Id: dl2pa.c,v 1.6 2002/05/02 08:09:31 dv Exp $
 *----------------------------------------------------------------------------
*/
# include <sys/types.h>
# include <sys/stat.h>
# include <sys/ipc.h>
# include <sys/shm.h>
# include <fcntl.h>
# include <signal.h>
# include <errno.h>
# include <stdlib.h>
# include <stdio.h>
# include <string.h>
# include <stdarg.h>
# include "lib/dl2pa.h"
# include "lib/monpa.h"
# include "lib/protnos.h"
# include "pa/pa.h"

# define COMMENT_MARK DL2_COMMENT_MARK
# define PACKETT_MARK DL2_PACKETT_MARK
# define PARAMST_MARK DL2_PARAMST_MARK
# define RCFILE DL2_RCFILE
# define ENVIRON DL2_ENVIRON

# define NSTR ""
# define NOT_IMPLEMENTED "not implemented"
# define BUFFSIZE 0x1000

# define OPT_USAGEM DL2_OPT_USAGEM
# define OPT_RCFILE DL2_OPT_RCFILE
# define OPT_FLAGS  DL2_OPT_FLAGS
# define OPT_INFILE "-i"
# define OPT_OUTFILE "-o"
# define OPT_FORMATI "-fi"
# define OPT_FORMATO "-fo"
# define OPT_SETFIELDS "-s"
# define OPT_ADDFIELDS "-a"
# define OPT_REMFIELDS "-r"
# define OPT_VT "-vt"
# define OPT_VF "-vf"

# define FORMATO_input 0
# define FORMATO_raw   1

# define KEYdefault "default"
# define KEYpacket_length "packet-length"
 
static int monfd=-1, shmid=-1, terminate=0;
static int verbose[PA_N_PTYPES];
static int verbose_set[PA_N_PTYPES];
static int verbose_add[PA_N_PTYPES];
static int verbose_rem[PA_N_PTYPES];
static const Pa_Packet* configuration[PA_N_PTYPES];
# define ADD_MODULE(pt) \
  configuration[PA_PID_##pt] = &pa_##pt;
static Dl2_Monpashm *shmmonpa;

static int options = 0, packet_options[PA_N_PTYPES];
# define SEL_RCFILE    0x00000002
# define SEL_INFILE    0x00000004
# define SEL_OUTFILE   0x00000008
# define SEL_FORMATI   0x00000010
# define SEL_FORMATO   0x00000020
# define SEL_SETFIELDS 0x00000040
# define SEL_ADDFIELDS 0x00000080
# define SEL_REMFIELDS 0x00000100
# define SEL_OFFLINE   0x80000000

static const char *dl2monrc;
static const char *infile, *outfile;
static const char *formati;
static int formato = FORMATO_input;
static int (*act_analyzer)(Dl2_Frame*,int);
static int act_verbose;
static Qoctet time_offset = -1;

static const char* locate_dl2monrc(void);
static const char* read_keyword(const char*,const char*); 
static int read_config(const char*);
static int read_config_cmdline(const char*,int*);
static const Pa_Packet* get_pconfig(const char*,int*);
static const Pa_Config* get_fconfig(const char*,const Pa_Packet*);
static int attach_shm(const char*);
static int nattach_shm(int);
static int detach_shm(void);
static char* str2literal(char*,char**);
static int set_analyzer(const Dl2_Monheader*);
static int read_packet(int,Dl2_Frame*);
static void message(char*,...);
static void print_usage(void);
static void on_terminate(int);


int main(int argc,char *argv[]) {
  int rc, i;
  int offline = 0;

  // initialize packet type configurations:
  ADD_MODULE(ether2);
  ADD_MODULE(pppoe);
  ADD_MODULE(lcp);
  ADD_MODULE(pap);
  ADD_MODULE(chap);
  ADD_MODULE(ipcp);
  ADD_MODULE(icmp);
  ADD_MODULE(tcp);
  ADD_MODULE(udp);
  ADD_MODULE(ip);

  // initialize library:
  dl2pa_libinit();

  // initialize verbose flags:
  for (i = 0; i < PA_N_PTYPES; i++) {
    verbose[i] = 0; 
    verbose_set[i] = 0; 
    verbose_add[i] = 0; 
    verbose_rem[i] = 0; 
    packet_options[i] = 0;
  }

  // read command line arguments:
  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i],OPT_USAGEM) == 0) {
      print_usage(); return 0;
    }

    else if (strcmp(argv[i],OPT_RCFILE) == 0 && i+1 < argc) {
      dl2monrc = argv[++i];
      options |= SEL_RCFILE;
    }

    else if (strcmp(argv[i],OPT_INFILE) == 0 && i+1 < argc) {
      infile = argv[++i];
      options |= SEL_INFILE;
    }

    else if (strcmp(argv[i],OPT_OUTFILE) == 0 && i+1 < argc) {
      outfile = argv[++i];
      options |= SEL_OUTFILE;
      message("%s: currently ignored\n",OPT_OUTFILE);
    }

    else if (strcmp(argv[i],OPT_FORMATI) == 0 && i+1 < argc) {
      formati = argv[++i];
      options |= SEL_FORMATI;
    }

    else if (strcmp(argv[i],OPT_FORMATO) == 0 && i+1 < argc) {
      const char *fostr = argv[++i];
      if (strcmp(fostr,"raw") == 0) {
        formato = FORMATO_raw;
      }
      else if (strcmp(fostr,"input") == 0) {
        formato = FORMATO_input;
      }
      else {
        message("%s: unknown output format '%s'\n",OPT_FORMATO,fostr);
        return -1;
      } options |= SEL_FORMATO;
    }

    else if (strncmp(argv[i],OPT_FLAGS,2) == 0) {
      if (strchr(argv[i],'o') != NULL) options |= SEL_OFFLINE;
    }

    else if (strcmp(argv[i],OPT_VT) == 0) {
      int j;
      for (j = 0; j < PA_N_PTYPES; j++) {
        const Pa_Packet *pc = configuration[j];
        const char *name;
        if (pc == NULL) continue;
        name = pc->name; 
        if (name != NULL && strlen(name) > 0) printf("%s ",name);
      } printf("\n"); 
      return 0;
    }

    else if (strcmp(argv[i],OPT_VF) == 0 && i+1 < argc) {
      const char *name = argv[++i];
      int idx;  
      const Pa_Packet* pc = get_pconfig(name,&idx);
      if (pc != NULL) {
        const Pa_Config *f;
        for (f = pc->field; f->cname != NULL; f++) {
          printf("%s(%s) ",f->cname,f->oname);
        } printf("\n");
      } return 0;
    }

    else if (strcmp(argv[i],OPT_SETFIELDS) == 0 && i+1 < argc) {
      const char *fields = argv[++i];
      int flags, pidx; 
      if ((pidx = read_config_cmdline(fields,&flags)) < 0) {
        message("error: %s %s\n",OPT_SETFIELDS,fields);
        return -1;
      } options |= SEL_SETFIELDS;
      verbose_set[pidx] = flags; 
      verbose_add[pidx] = verbose_rem[pidx] = 0; 
      packet_options[pidx] |= SEL_SETFIELDS;
    }

    else if (strcmp(argv[i],OPT_ADDFIELDS) == 0 && i+1 < argc) {
      const char *fields = argv[++i];
      int flags, pidx; 
      if ((pidx = read_config_cmdline(fields,&flags)) < 0) {
        message("error: %s %s\n",OPT_ADDFIELDS,fields);
        return -1;
      } options |= SEL_ADDFIELDS;
      verbose_add[pidx] |= flags; packet_options[pidx] |= SEL_ADDFIELDS;
    }

    else if (strcmp(argv[i],OPT_REMFIELDS) == 0 && i+1 < argc) {
      const char *fields = argv[++i];
      int flags, pidx; 
      if ((pidx = read_config_cmdline(fields,&flags)) < 0) {
        message("error: %s %s\n",OPT_REMFIELDS,fields);
        return -1;
      } options |= SEL_REMFIELDS;
      verbose_rem[pidx] |= flags; packet_options[pidx] |= SEL_REMFIELDS;
    }

    else {
      message("%s: unknown option\n",argv[i]);
      return -1;
    }
  }

  // check command line arguments:
  if ((options & SEL_RCFILE) == 0) {
    dl2monrc = locate_dl2monrc();
  }
  if (dl2monrc != NULL) {
    if (read_config(dl2monrc) < 0) {
      message("error: reading config from %s\n",dl2monrc);
      return -1;
    } options |= SEL_RCFILE;
  }

  // refine verbose flags by command line:
  for (i = 0; i < PA_N_PTYPES; i++) {
    if (packet_options[i] & SEL_SETFIELDS) {
      verbose[i] = verbose_set[i];
    }
    if (packet_options[i] & SEL_ADDFIELDS) {
      verbose[i] |= verbose_add[i];
    }
    if (packet_options[i] & SEL_REMFIELDS) {
      verbose[i] &= ~verbose_rem[i];
    }
  }

  if ((options & SEL_INFILE) == 0) {
    if ((options & SEL_RCFILE) == 0) {
      message("error: missing infile\n");
      return -1;
    }
    infile = read_keyword(dl2monrc,DL2_KEY_MONOUT);
    if (infile == NULL) {
      message("error: reading infile from %s\n",dl2monrc);
      return -1;
    } options |= SEL_INFILE;
  }

  // check for offline, test infile:
  offline = (options & SEL_OFFLINE);
  if (offline && ((options & SEL_INFILE)==0)) {
    message("error: offline requires infile\n");
    return -1;
  }
 
  // outfile: tbd
  // field options: tbd
  // formati: tbd

  // initialize shared memory:
  if (offline == 0) {
    if (attach_shm(DL2_SHMPATHNAME) < 0) {
      message("error: attaching shm %s\n",DL2_SHMPATHNAME);
      return -1;
    }
  }

  // install signal handlers:
  if (offline) {
    signal(SIGTERM,on_terminate);
    signal(SIGINT,on_terminate);
  }

  // open the input file (online: stdin):
  if (offline) {
    monfd = (rc = open(infile,O_RDONLY|O_ASYNC));
  } else monfd = 0;

  // answer here-we-are:
  if (offline == 0) {
    shmmonpa->flags |= DL2_SHMF_paup;
  }
 
  if (monfd >= 0) {
    Dl2_Frame frame;
    Dl2_Monheader header;

    message("dl2pa started\n");
    rc = read(monfd,&header,sizeof(header));
    if (rc >= sizeof(header)) {
      if (formato == FORMATO_raw) {
        act_analyzer = dl2_paraw;
        act_verbose = 0;
      }
      else rc = set_analyzer(&header);
    } else goto main_terminate;

    if (offline) {
      while (terminate == 0 && rc >= 0) {
        rc = read_packet(monfd,&frame);
        if (rc == 0) break;
      }
    }
    else {
      while (terminate == 0 && rc >= 0) {
        rc = read_packet(monfd,&frame);
        if (rc == 0) sleep(1);
      }
    }
    close(monfd); monfd = -1;

  main_terminate:
    on_terminate(-1);
    message("dl2pa termed (%d)\n",rc);
    return rc;
  }
  else {
    if (shmmonpa != NULL) detach_shm();
  } return -1;
} //main


int pa_verbose(int pid) {
  if (0 <= pid < PA_N_PTYPES) {
    return verbose[pid];
  } return 0;
} //pa_verbose

 
//-locals---------------------------------------------------------------------
/* search for .dl2monrc, write absolute pathname to s; return -1 on error;
 * search order: environment, HOME.
*/
const char* locate_dl2monrc(void) {
  const char *dl2mondir = getenv(ENVIRON);
  if (dl2mondir == NULL) {
    dl2mondir = getenv("HOME");
  }
  if (dl2mondir != NULL) {
    char *s = malloc(strlen(dl2mondir)+strlen(RCFILE)+2);
    sprintf(s,"%s/%s",dl2mondir,RCFILE);
    return s;
  } return NULL;  
} //locate_dl2monrc


/* search rcfile cf for keyword key;
 * return its value or NSTR, NULL on error;
*/
const char* read_keyword(const char *cf,const char *key) {
  FILE *f = fopen(cf,"r");
  char *value = NULL;
  if (f != NULL) {
    char line[512], *p, *q;
    while (fgets(line,512,f) != NULL) {
      p = str2literal(line,NULL);
      if (p[0] == COMMENT_MARK || p[0] == 0) continue;
      if ((q = strchr(p,PARAMST_MARK)) == NULL) continue;
      *q = 0;
      if (strcmp(p,key) == 0) {
        p = str2literal(q+1,&q);         
        if (q > p) {
          value = malloc(q-p+1);
          strncpy(value,p,q-p);  
        } else value = NSTR; 
      }
    }
    if (ferror(f) != 0) {
      if (value != NULL) free(value);
      value = NULL;
    }    
    fclose(f);
  } return value;
} //read_keyword
 

int read_config(const char *cf) {
  FILE *f = fopen(cf,"r");
  int rc = 0;
  if (f != NULL) {
    char line[512], *p, *q, *maxp;
    while (fgets(line,512,f) != NULL) {
      const Pa_Packet *packet;
      const Pa_Config *field;
      int pidx;
      maxp = line + strlen(line);
      p = str2literal(line,NULL);
      if (p[0] == COMMENT_MARK || p[0] == 0) continue;
      if ((q = strchr(p,PACKETT_MARK)) == NULL) continue;
      *q = 0;
      if ((packet = get_pconfig(p,&pidx)) == NULL) {rc = -1;break;}
      for (p = q+1; p < maxp; p = q+1) {
        p = str2literal(p,&q);
        if (q > p) {
          *q = 0;
          // reserved verbose flags are handled here:
          if (strcmp(p,KEYdefault) == 0) {
            verbose[pidx] |= PA_VMdefault;
            continue;
          }
          else if (strcmp(p,KEYpacket_length) == 0) {
            verbose[pidx] |= PA_VMpacket_length;
            continue;
          }
          // others than reserved:
          if ((field = get_fconfig(p,packet)) == NULL) {rc = -1;break;}
          verbose[pidx] |= (1<<field->index); 
        } else break;
      }
    }
    if (ferror(f) != 0) rc = -1;    
    fclose(f);
    return rc;
  } return -1;
} //read_config


/* The same function as read_config, but with the argument 'spec' 
 * representing one specification for field selection; the selected 
 * verbose flags will be returned by 'verbose';
 * return -1 on error, packet index otherwise;
*/
int read_config_cmdline(const char *spec, int *verbose) {
  char line[1024], *p, *q, *maxp;
  const Pa_Packet *packet;
  const Pa_Config *field;
  int pidx;
  *verbose = 0;
  if (strlen(spec) > 1024) return -1;
  strcpy(line,spec); 
  maxp = line + strlen(line);
  p = str2literal(line,NULL);
  if ((q = strchr(p,PACKETT_MARK)) == NULL) return -1;
  *q = 0;
  if ((packet = get_pconfig(p,&pidx)) == NULL) return -1;
  for (p = q+1; p < maxp; p = q+1) {
    p = str2literal(p,&q);
    if (q > p) {
      *q = 0;
      // reserved verbose flags are handled here:
      if (strcmp(p,KEYdefault) == 0) {
        *verbose |= PA_VMdefault;
        continue;
      }
      else if (strcmp(p,KEYpacket_length) == 0) {
        *verbose |= PA_VMpacket_length;
        continue;
      }
      // others than reserved:
      if ((field = get_fconfig(p,packet)) == NULL) return -1;
      *verbose |= (1<<field->index); 
    } else break;
  }
  return pidx;
} //read_config_cmdline


const Pa_Packet* get_pconfig(const char *name,int *index) {
  int i;
  for (i = 0; i < PA_N_PTYPES; i++) {
    if (configuration[i] == NULL) continue;
    if (strcmp(name,configuration[i]->name) == 0) {
      *index = i;
      return configuration[i];
    }
  } return NULL;    
} //get_pconfig


const Pa_Config* get_fconfig(const char *cname,const Pa_Packet *pc) {
  const Pa_Config *f;
  for (f = pc->field; f->cname != NULL; f++) {
    if (strcmp(cname,f->cname) == 0) return f;
  } return NULL; 
} //get_fconfig


char* str2literal(char* p,char** q) {
  char *r = p + strspn(p," \t\n,"); 
  if (q != NULL) *q = r + strcspn(r," \t\n,"); 
  return r;
} //str2literal


//-locals---------------------------------------------------------------------
/* attach to the memory shared with the DL2 Monitor: set shmid, shmmonpa;
 * the process doesn't create shm, but attaches in RW mode; on success, it
 * the pid to the shm data; the number of trials has been restricted to 10; 
 * return -1 on error;
*/
int attach_shm(const char *pn) {
  int trials, rc;
  key_t key = ftok(pn,1);
  if (key < 0) return -1;
  for (trials = 0; trials < 10; trials++) {
    shmid = shmget(key,0,DL2_SHMV_MODE);
    if (shmid >= 0) {
      struct shmid_ds ds;
      if ((rc = shmctl(shmid,IPC_STAT,&ds)) < 0) break;
      if (ds.shm_nattch > 1) break;
      shmmonpa = (Dl2_Monpashm*)shmat(shmid,NULL,0);
      if ((int)shmmonpa == -1) {shmmonpa = NULL; break;}
      if (shmmonpa->flags & DL2_SHMF_paup) {
        shmdt(shmmonpa); break;
      }
      shmmonpa->papid = getpid();
      return 0;
    } else sleep(5);
  } return -1;
} //attach_shm


/* detach from the shm; if we are the final attached process,
 * then remove the shm from system; always return 0;
*/
int detach_shm(void) {
  struct shmid_ds ds;
  int rc = shmctl(shmid,IPC_STAT,&ds);
  if (rc >= 0 && (ds.shm_perm.uid == geteuid())) {
    shmmonpa->flags &= ~DL2_SHMF_paup;
    shmdt(shmmonpa);
    if (ds.shm_nattch == 1) shmctl(shmid,IPC_RMID,NULL);
  }
  shmmonpa = NULL; shmid = -1;
  return 0;
} //detach_shm


int nattach_shm(int id) {
  struct shmid_ds ds;
  int rc = shmctl(id,IPC_STAT,&ds);
  return (rc >= 0 ? ds.shm_nattch : rc);
} //nattach_shm 


//-locals---------------------------------------------------------------------
/* hack:
 * select an analyzer with respect to the PPPoE capabilities:
 * choice depends on socket-type, protocol, and perhaps interface;
 * protocol       raw        datagram  |  ppp?
 * DL2_PD_all     paether2   papppoe   |  paip
 * DL2_PT_PPP_*   paether2   papppoe   |  paip
 * DL2_PT_ip      paip       paip      |  paip
 * In cases other than DL2_PT_ip, some ip packets will be lost. 
*/
int set_analyzer(const Dl2_Monheader *hdr) {
# define digits "0123456789"
  if (hdr->socket_type == DL2_ST_raw) {
    act_analyzer = dl2_paether2;
    act_verbose = verbose[PA_PID_ether2];
  }
  else if (hdr->socket_type == DL2_ST_dgram) {
    act_analyzer = dl2_papppoe;
    act_verbose = verbose[PA_PID_pppoe];
  }
  else return -1;
  if (hdr->protocol == DL2_PT_ip) {
    act_analyzer = dl2_paip;
    act_verbose = verbose[PA_PID_ip];
  }
  // I don't why, but it works in some cases!
  else if (strncmp(hdr->interface,"ppp",3) == 0 &&
           strchr(digits,hdr->interface[3]) != NULL) {
    act_analyzer = dl2_paip;
    act_verbose = verbose[PA_PID_ip];
  } return 1;
# undef digits
} //set_analyzer


/* the function analyzes one packet per call and writes the result to
 * stdout; return the number of processed octets, -1 on error;
 * the analyzer in effect is kept from static act_analyzer, act_verbose;
 * if reading timestamp or packet length return 0, the function returns
 * immediatly instead of looping on the read() operation; this is because
 * of reading 0 bytes is considered legal from the system side.
*/
int read_packet(int fd,Dl2_Frame *frame) {
  Octet buffer[BUFFSIZE];
  Qoctet timestamp, difftime;
  int rc, length, n, nr = 0;

  // read timestamp of the packet:
  length = DL2_TIME_SIZE; n = 0;
  do {
    n += (rc = read(fd,buffer+n,length-n));
    if (rc <= 0) return rc;
  } while (n < length); nr += length;

  // convert timestamp:
  timestamp = str2qoct(buffer);
  if (timestamp < time_offset) time_offset = timestamp;
  difftime = timestamp-time_offset;
  printf("%06u ",difftime);

  // read length of the packet:
  length = DL2_LENGTH_SIZE; n = 0;
  do {
    n += (rc = read(fd,buffer+n,length-n));
    if (rc <= 0) return rc;
  } while (n < length); nr += length;

  // print out the packet length if requested:
  length = str2doct(buffer);
  if (act_verbose & PA_VMpacket_length) {
    printf("%s=%04x ",PA_VFpacket_length,length);
  }

  // read data of the packet:
  if (length == 0) return nr;
  n = 0;
  do {
    n += (rc = read(fd,buffer+n,length-n));
    if (rc < 0) return rc;
  } while (n < length); nr += length;

  // call protocol/packet analyzer:
  frame->buffer = buffer;
  frame->length = length;
  if (act_analyzer(frame,act_verbose) < 0) printf("<unknown>");
  printf("\n"); 
  fflush(stdout);

  return nr;
} //read_packet


//-locals---------------------------------------------------------------------
void message(char *fmt,...) {
  char sfmt[256];
  va_list ap;
  va_start(ap,fmt);
  sprintf(sfmt,"**%s",fmt);
  vfprintf(stderr,sfmt,ap);
  va_end(ap);  
} //message


void print_usage(void) {
  printf("DL2 Monitor (analyzer), Version %.2lf (%s)\n",
    DL2_VERSION_NO,DL2_VERSION_DATE);
  printf("usage: dl2pa {<analyzer-option>}\n");
  printf("analyzer-option (default):\n\
%s <file> use <file> in place of the rcfile (%s)\n\
%s <file> read monitor input from <file> (%s), offline\n\
%s <ifmt> expect input according to format <ifmt> (%s), offline\n\
%s <p>:{<f>} set field selection for packet <p> to {<f>} (%s)\n\
%s <p>:{<f>} add field selection {<f>} for packet <p> (%s)\n\
%s <p>:{<f>} remove field selection {<f>} for packet <p> (%s)\n\
%s <file> write output to <file> instead of stdout (%s)\n\
%s <ofmt> generate output according to format <ofmt> (%s)\n\
%s print known packet types and exit (%s)\n\
%s <vt> printf known fields for packet type <vt> and exit (%s)\n\
%s{<flag>} (%s)\n\
",
OPT_RCFILE,".dl2monrc",
OPT_INFILE,"rcfile.mon-output=",
OPT_FORMATI,NOT_IMPLEMENTED,
OPT_SETFIELDS,"rcfile",
OPT_ADDFIELDS,"",
OPT_REMFIELDS,"",
OPT_OUTFILE,NOT_IMPLEMENTED,
OPT_FORMATO,"input",
OPT_VT,"",
OPT_VF,"",
OPT_FLAGS,"");
} //print_usage


// invoke this procedure only once!
void on_terminate(int sig) {
  terminate = 1; 
  if (monfd > 0) close(monfd);
  if (shmmonpa != NULL) detach_shm();
  if (sig != -1) {
    fprintf(stderr,"**dl2pa signal-termed (%d)\n",sig);
    exit(0);
  }
} //on_terminate

/***EOM***/
